"""
This module contains the base :class:`.EngineerException` .
"""


class EngineerException(Exception):
    """
    The base class for errors in :mod:`royalnet.engineer`.
    """


__all__ = (
    "EngineerException",
)
